unit TetrisGameClass;

interface

uses
  Classes, Graphics, SysUtils, Types, StdCtrls, Grids, GameGridClass, ExtCtrls;

const
  clFixed = clBtnFace;
  clEmpty = clWhite;
  clFigure = clBlue;
  clDropped = clNavy;


type
  TTetrisGame = class (TGameGrid)
  private
    fOnDeleteRow: TNotifyEvent;
    fOnFailed: TNotifyEvent;
    procedure Freeze;
    procedure NewFigure;
    function CheckSpace (dx, dy: integer): boolean;
    procedure CheckState;
  protected
    procedure DeleteRow (ARow: integer); override;
  public
    constructor Create (AComponent: TComponent); override;
    procedure DrawCell (ACol, ARow: integer;
      ARect: TRect; AState: TGridDrawState); override;
    procedure FigureLeft;
    procedure FigureRight;
    procedure Move; override;
    procedure MoveFigure (dx, dy: integer);
    procedure Drop;
    procedure Paint; override;
    procedure Prepare; override;
    procedure Rotate;
    property OnDeleteRow: TNotifyEvent read fOnDeleteRow write fOnDeleteRow;
    property OnFailed: TNotifyEvent read fOnFailed write fOnFailed;
  end;

implementation

{ TTetrisGame }

function TTetrisGame.CheckSpace(dx, dy: integer): boolean;
var
  i, j: integer;
begin
  Result := TRUE;
  for j := RowCount - dy - 1 downto 0 do begin
    for i := 1 to ColCount - 2 do begin
      if Cells [i, j] = '1' then begin
        if (Cells [i + dx, j + dy] = '0') or
          (Cells [i + dx, j + dy] = '2') then
          Result := FALSE;
      end;
    end;
  end;
end;

procedure TTetrisGame.CheckState;
var
  i, j, n: integer;
  AFilledRow: boolean;
begin
  n := 1;
  j := RowCount - 2;
  while j >= 0 do begin
    AFilledRow := TRUE;
    for i := 1 to ColCount - 2 do
      if Cells [i, j] <> '2' then
        AFilledRow := FALSE;
    if AFilledRow then begin
      DeleteRow (j);
      Score := Score + n;
      n := n * 2;
      Speed := Speed - n * 10;
      if Assigned (fOnDeleteRow) then
        fOnDeleteRow (Self);
    end else
      Dec (j);
  end;

  for j := RowCount - 1 downto 0 do
    for i := 1 to ColCount - 2 do
      if (Cells [i, j] = '2') and (j < 4) then begin
        State := gsUserLoose;
        if Assigned (fOnFailed) then
          fOnFailed (Self);
        Exit;
      end;
end;

constructor TTetrisGame.Create(AComponent: TComponent);
begin
  inherited;
  FixedCols := 0;
  FixedRows := 0;
  ColCount := 11;
  RowCount := 21;
  DefaultColWidth := 16;
  DefaultRowHeight := 16;
  DefaultDrawing := FALSE;
  GridLineWidth := 0;
  ScrollBars := ssNone;
  Options := [];
  Width := (DefaultColWidth + GridLineWidth) * ColCount + 4;
  Height := (DefaultRowHeight + GridLineWidth) * RowCount + 4;
  Prepare;
  Randomize;
end;


procedure TTetrisGame.DeleteRow(ARow: integer);
var
  i, j: integer;
begin
  for j := ARow downto 1 do
    for i := 1 to ColCount - 2 do
      Cells [i, j] := Cells [i, j - 1];
  for i := 1 to ColCount - 2 do
    Cells [i, 0] := '';
end;

procedure TTetrisGame.DrawCell(ACol, ARow: integer; ARect: TRect;
  AState: TGridDrawState);
var
  AValue: integer;
begin
  inherited;
  with Canvas do begin
    if Cells [ACol, ARow] <> '' then begin
      AValue := StrToInt (Cells [ACol, ARow]);
      case AValue of
        0: begin
          Brush.Color := clFixed;
          FillRect (ARect);
          Frame3D (Canvas, ARect, clWhite, clGray, 1);
        end;
        1: begin
          Brush.Color := clFigure;
          FillRect (ARect);
          Frame3D (Canvas, ARect, clAqua, clNavy, 1);
        end;
        2: begin
          Brush.Color := clDropped;
          FillRect (ARect);
          Frame3D (Canvas, ARect, clBlue, clBlack, 1);
        end;
      end;
    end else begin
      Brush.Color := clEmpty;
      FillRect (ARect);
    end;
  end;

end;

procedure TTetrisGame.Drop;
begin
  if CheckSpace (0, 1) then
    MoveFigure (0, 1);
end;

procedure TTetrisGame.FigureLeft;
begin
  if CheckSpace (-1, 0) then
    MoveFigure (-1, 0);
end;

procedure TTetrisGame.FigureRight;
begin
  if CheckSpace (1, 0) then
    MoveFigure (1, 0);
end;

procedure TTetrisGame.Freeze;
var
  i, j: integer;
begin
  for j := 0 to RowCount - 1 do
    for i := 1 to ColCount - 2 do
      if Cells [i, j] = '1' then
        Cells [i, j] := '2';
end;

procedure TTetrisGame.Move;
begin
  if CheckSpace (0, 1) then
    MoveFigure (0, 1)
  else begin
    Freeze;
    CheckState;
    if State = gsPlaying then
      NewFigure;
  end;
end;

procedure TTetrisGame.MoveFigure(dx, dy: integer);
var
  i, j: integer;
begin
  for j := RowCount - 2 downto 0 do
    if dx < 0 then begin
      for i := 1 to ColCount - 2 do
        if Cells [i, j] = '1' then
          if Cells [i + dx, j + dy] = '' then begin
            Cells [i + dx, j + dy] := Cells [i, j];
            Cells [i, j] := '';
          end;
    end else begin
      for i := ColCount - 2 downto 1 do
        if Cells [i, j] = '1' then
          if Cells [i + dx, j + dy] = '' then begin
            Cells [i + dx, j + dy] := Cells [i, j];
            Cells [i, j] := '';
          end;
    end;
end;

procedure TTetrisGame.NewFigure;
var
  AFigure,
  fLeft,
  fTop: integer;
begin
  AFigure := Random (15);
  fLeft := Random (ColCount - 6) + 1;
  fTop := 0;

  case AFigure of
    0: begin
      Cells [fLeft + 2, fTop] := '1';
      Cells [fLeft + 2, fTop + 1] := '1';
      Cells [fLeft + 2, fTop + 2] := '1';
      Cells [fLeft + 2, fTop + 3] := '1';
    end;
    1: begin
      Cells [fLeft + 1, fTop] := '1';
      Cells [fLeft + 2, fTop] := '1';
      Cells [fLeft + 1, fTop + 1] := '1';
      Cells [fLeft + 2, fTop + 1] := '1';
    end;
    2: begin
      Cells [fLeft + 1, fTop] := '1';
      Cells [fLeft + 2, fTop] := '1';
      Cells [fLeft + 2, fTop + 1] := '1';
      Cells [fLeft + 3, fTop + 1] := '1';
    end;
    3: begin
      Cells [fLeft + 2, fTop] := '1';
      Cells [fLeft + 3, fTop] := '1';
      Cells [fLeft + 1, fTop + 1] := '1';
      Cells [fLeft + 2, fTop + 1] := '1';
    end;
    4: begin
      Cells [fLeft + 1, fTop] := '1';
      Cells [fLeft + 2, fTop] := '1';
      Cells [fLeft + 3, fTop] := '1';
      Cells [fLeft + 2, fTop + 1] := '1';
    end;
    5: begin
      Cells [fLeft, fTop] := '1';
      Cells [fLeft, fTop + 1] := '1';
      Cells [fLeft, fTop + 2] := '1';
      Cells [fLeft, fTop + 3] := '1';
    end;
    6: begin
      Cells [fLeft, fTop] := '1';
      Cells [fLeft, fTop + 1] := '1';
      Cells [fLeft + 1, fTop + 1] := '1';
      Cells [fLeft + 1, fTop + 2] := '1';
    end;
    7: begin
      Cells [fLeft + 1, fTop] := '1';
      Cells [fLeft + 1, fTop + 1] := '1';
      Cells [fLeft, fTop + 1] := '1';
      Cells [fLeft, fTop + 2] := '1';
    end;
    8: begin
      Cells [fLeft, fTop + 1] := '1';
      Cells [fLeft + 1, fTop + 1] := '1';
      Cells [fLeft + 2, fTop + 1] := '1';
      Cells [fLeft + 1, fTop] := '1';
    end;
    9: begin
      Cells [fLeft, fTop] := '1';
      Cells [fLeft + 1, fTop] := '1';
      Cells [fLeft + 1, fTop + 1] := '1';
      Cells [fLeft + 1, fTop + 2] := '1';
    end;
    10: begin
      Cells [fLeft, fTop] := '1';
      Cells [fLeft + 1, fTop] := '1';
      Cells [fLeft, fTop + 1] := '1';
      Cells [fLeft, fTop + 2] := '1';
    end;
    11: begin
      Cells [fLeft, fTop] := '1';
      Cells [fLeft + 1, fTop] := '1';
      Cells [fLeft + 2, fTop] := '1';
      Cells [fLeft, fTop + 1] := '1';
    end;
    12: begin
      Cells [fLeft, fTop] := '1';
      Cells [fLeft + 1, fTop] := '1';
      Cells [fLeft + 2, fTop] := '1';
      Cells [fLeft + 2, fTop + 1] := '1';
    end;
    13: begin
      Cells [fLeft, fTop] := '1';
      Cells [fLeft, fTop + 1] := '1';
      Cells [fLeft, fTop + 2] := '1';
      Cells [fLeft + 1, fTop + 1] := '1';
    end;
    14: begin
      Cells [fLeft + 1, fTop] := '1';
      Cells [fLeft + 1, fTop + 1] := '1';
      Cells [fLeft + 1, fTop + 2] := '1';
      Cells [fLeft, fTop + 1] := '1';
    end;
  end;
end;

procedure TTetrisGame.Paint;
begin
  inherited;
  with Canvas do begin
    Pen.Color := clRed;
    MoveTo (0, 4 * DefaultRowHeight - 1);
    LineTo (Width, 4 * DefaultRowHeight - 1);
  end;
end;

procedure TTetrisGame.Prepare;
var
  i, j: integer;
begin
  for i := 1 to ColCount - 2 do
    for j := 0 to RowCount - 2 do
      Cells [i, j] := '';
  for i := 0 to RowCount - 1 do begin
    Cells [0, i] := '0';
    Cells [ColCount - 1, i] := '0';
  end;
  for i := 1 to ColCount - 2 do
    Cells [i, RowCount - 1] := '0';
  NewFigure;
end;

procedure TTetrisGame.Rotate;
begin
end;

end.
